'''
Python module that contains the functions
about reading and writing files.
'''
from __future__ import print_function

import os
from pathlib import Path
import numpy as _np
from scipy import sparse as sp
from scipy.sparse import csgraph as cs
from tmd.io.swc import SWC_DCT
from tmd.io.swc import read_swc
from tmd.io.swc import swc_to_data
from tmd.io.h5 import read_h5
from tmd.io.conversion import convert_morphio_soma
from tmd.io.conversion import convert_morphio_trees
from tmd.Neuron import Neuron
from tmd.Tree import Tree
from tmd.Soma import Soma
from tmd.Population import Population
from tmd.utils import TREE_TYPE_DICT
from tmd.utils import SOMA_TYPE


class LoadNeuronError(Exception):
    '''Captures the exception of failing to load a single neuron
    '''


def make_tree(data):
    '''Make tree structure from loaded data.
       Returns a tree of tmd.Tree
       type.
    '''
    tr_data = _np.transpose(data)

    parents = [_np.where(tr_data[0] == i)[0][0]
               if len(_np.where(tr_data[0] == i)[0]) > 0
               else - 1 for i in tr_data[6]]

    return Tree.Tree(x=tr_data[SWC_DCT['x']], y=tr_data[SWC_DCT['y']], z=tr_data[SWC_DCT['z']],
                     d=tr_data[SWC_DCT['radius']], t=tr_data[SWC_DCT['type']], p=parents)


def redefine_types(user_types=None):
    """
    Returns tree types depending on the customized types
        selected by the user.

        Args:
            user_types (dictionary or None):

        Returns:
            final_types (dict): tree types for the construction of Neuron.
    """
    final_tree_types = TREE_TYPE_DICT.copy()
    if user_types is not None:
        final_tree_types.update(user_types)
    return final_tree_types


def load_neuron(input_file, line_delimiter='\n', soma_type=None,
                user_tree_types=None, remove_duplicates=True):
    """
    Io method to load an swc or h5 file into a Neuron object.
    """
    tree_types = redefine_types(user_tree_types)

    # Definition of swc types from type_dict function
    if soma_type is None:
        soma_index = SOMA_TYPE
    else:
        soma_index = soma_type

    # Make neuron with correct filename and load data
    if os.path.splitext(input_file)[-1] == '.swc':
        data = swc_to_data(read_swc(input_file=input_file,
                                    line_delimiter=line_delimiter))
        neuron = Neuron.Neuron(name=input_file.replace('.swc', ''))

    elif os.path.splitext(input_file)[-1] == '.h5':
        data = read_h5(input_file=input_file, remove_duplicates=remove_duplicates)
        neuron = Neuron.Neuron(name=input_file.replace('.h5', ''))

    try:
        soma_ids = _np.where(_np.transpose(data)[1] == soma_index)[0]
    except IndexError as exc:
        raise LoadNeuronError('Soma points not in the expected format') from exc

    # Extract soma information from swc
    soma = Soma.Soma(x=_np.transpose(data)[SWC_DCT['x']][soma_ids],
                     y=_np.transpose(data)[SWC_DCT['y']][soma_ids],
                     z=_np.transpose(data)[SWC_DCT['z']][soma_ids],
                     d=_np.transpose(data)[SWC_DCT['radius']][soma_ids])

    # Save soma in Neuron
    neuron.set_soma(soma)
    p = _np.array(_np.transpose(data)[6], dtype=int) - _np.transpose(data)[0][0]
    # return p, soma_ids
    try:
        dA = sp.csr_matrix((_np.ones(len(p) - len(soma_ids)),
                           (range(len(soma_ids), len(p)),
                            p[len(soma_ids):])), shape=(len(p), len(p)))
    except Exception as exc:
        raise LoadNeuronError('Cannot create connectivity, nodes not connected correctly.') from exc

    # assuming soma points are in the beginning of the file.
    comp = cs.connected_components(dA[len(soma_ids):, len(soma_ids):])

    # Extract trees
    for i in range(comp[0]):
        tree_ids = _np.where(comp[1] == i)[0] + len(soma_ids)
        tree = make_tree(data[tree_ids])
        neuron.append_tree(tree, tree_types)

    return neuron


def load_neuron_from_morphio(path_or_obj, user_tree_types=None):
    """
    Create Neuron object from morphio object or from path
        loaded via morphio.
        Supported file formats: h5, swc, asc.

        Args:
            path_or_obj (Union[str, morphio.Morphology]):
                Filepath or morphio object

        Returns:
            neuron (Neuron): tmd Neuron object
    """
    from morphio import Morphology  # pylint: disable=import-outside-toplevel

    tree_types = redefine_types(user_tree_types)

    if isinstance(path_or_obj, (str, Path)):
        obj = Morphology(path_or_obj)
        filename = path_or_obj
    else:
        obj = path_or_obj
        # MorphIO does not support naming of objects yet.
        filename = ''

    neuron = Neuron.Neuron()
    neuron.name = filename
    neuron.set_soma(convert_morphio_soma(obj.soma))
    for tree in convert_morphio_trees(obj):
        neuron.append_tree(tree, tree_types)

    return neuron


def load_population(neurons, user_tree_types=None, name=None, use_morphio=False):
    '''Loads all data of recognised format (swc, h5)
       into a Population object.
       Takes as input a directory or a list of files to load.
    '''
    if isinstance(neurons, (list, tuple)):
        files = neurons
        name = name if name is not None else 'Population'
    elif os.path.isdir(neurons):  # Assumes given input is a directory
        files = [os.path.join(neurons, neuron_dir) for neuron_dir in os.listdir(neurons)]
        name = name if name is not None else os.path.basename(neurons)

    pop = Population.Population(name=name)

    for filename in files:
        try:
            if not use_morphio:
                assert filename.endswith(('.h5', '.swc'))
                pop.append_neuron(load_neuron(filename, user_tree_types=user_tree_types))
            else:
                assert filename.endswith(('.h5', '.swc', '.asc'))
                pop.append_neuron(load_neuron_from_morphio(filename,
                                                           user_tree_types=user_tree_types))

        except AssertionError as exc:
            error_msg = "{} is not a valid h5, swc or asc file. If asc set use_morphio to True."
            raise Warning(error_msg.format(filename)) from exc
        except LoadNeuronError:
            print('File failed to load: {}'.format(filename))

    return pop



from mock import patch, Mock
import numpy as np
from numpy import testing as npt
from tmd.Topology import persistent_properties as tested


class MockTree:

    def __init__(self, points, parents=None, children=None, begs=None, ends=None):

        self._points = points
        self._parents = parents
        self._children = children
        self._begs = begs
        self._ends = ends

    def get_direction_between(self, start_id, end_id):
        vec = self._points[end_id] - self._points[start_id]
        vec /= np.linalg.norm(vec)
        return vec

    @property
    def sections(self):
        return self._begs, self._ends

    @property
    def parents_children(self):
        return self._parents, self._children


def test_no_property():

    prop = tested.NoProperty(None)

    res_get = prop.get(None)
    res_inf = prop.infinite_component(None)

    assert isinstance(res_get, list)
    assert isinstance(res_inf, list)

    npt.assert_equal(len(res_get), 0)
    npt.assert_equal(len(res_inf), 0)


def test_persistent_angles():

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [0., 0., -1.],
            [0., 1., 0.]
        ]),
        parents={1: 0},
        children={1: [2, 3]},
        begs=[-1, 1]
    )

    prop = tested.PersistentAngles(tree)

    res_get = prop.get(1)
    res_inf = prop.infinite_component(0)

    assert isinstance(res_get, list)
    assert isinstance(res_inf, list)

    npt.assert_allclose(res_get, [0., 0., 0.5 * np.pi, -0.5 * np.pi])
    npt.assert_array_equal(res_inf, [np.nan, np.nan, np.nan, np.nan])


def test_section_mean_radii():

    radii = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])

    section_begs = np.array([0, 2, 5])
    section_ends = np.array([2, 5, 6])

    expected_mean_radii = [0.15, 0.4, 0.6]

    npt.assert_allclose(
        tested.PersistentMeanRadius._section_mean_radii(radii, section_begs, section_ends),
        expected_mean_radii
    )


def test_persistent_mean_radius():

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [0., 0., -1.],
            [0., 1., 0.]
        ]),
        begs=np.array([0, 1, 2, 3]),
        ends=np.array([1, 2, 3, 4])
    )
    tree.d = 2.0 * np.array([0.1, 0.2, 0.3, 0.4])

    prop = tested.PersistentMeanRadius(tree)

    res_get = prop.get(2)
    res_inf = prop.infinite_component(0)

    assert isinstance(res_get, list)
    assert isinstance(res_inf, list)

    npt.assert_array_equal(res_get, [0.3])
    npt.assert_array_equal(res_inf, [0.1])


def test_phi_theta():

    func = tested.PersistentAngles._phi_theta

    u = np.array([1., 0., 0.])
    v = np.array([1., 0., 0.])

    npt.assert_allclose(func(u, v), [0., 0.])
    npt.assert_allclose(func(v, u), [0., 0.])
    npt.assert_allclose(func(3. * u, 4. * v), [0., 0.])

    u = np.array([1., 0., 0.])
    v = np.array([0., 0., 1.])

    npt.assert_allclose(func(u, v), [0., -np.pi * 0.5])
    npt.assert_allclose(func(v, u), [0., +np.pi * 0.5])
    npt.assert_allclose(func(2. * v, 3. * u), [0., np.pi * 0.5])

    u = np.array([1., 0., 0.])
    v = np.array([-1., 0., 0.])

    npt.assert_allclose(func(u, v), [np.pi, 0.])
    npt.assert_allclose(func(v, u), [-np.pi, 0.])
    npt.assert_allclose(func(2. * v, 3. * u), [-np.pi, 0.])

    u = np.array([1., 1., 0.])
    v = np.array([1., 0., 1.])

    npt.assert_allclose(func(u, v), [-0.25 * np.pi, -0.25 * np.pi])
    npt.assert_allclose(func(v, u), [0.25 * np.pi, 0.25 * np.pi])
    npt.assert_allclose(func(2. * v, 3. * u), [0.25 * np.pi, 0.25 * np.pi])


def test_angles_tree():

    func = tested.PersistentAngles._angles_tree

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [1., 0., 0.],
            [0., 1., 0.]
        ])
    )

    angles = func(tree, parID=0, parEND=1, ch1ID=2, ch2ID=3)
    npt.assert_allclose(angles, [0., -0.5 * np.pi, 0.5 * np.pi, 0.])

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [0., 0., -1.],
            [0., 1., 0.]
        ])
    )

    angles = func(tree, parID=0, parEND=1, ch1ID=2, ch2ID=3)
    npt.assert_allclose(angles, [0., 0., 0.5 * np.pi, -0.5 * np.pi])

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [0., 1., 0.],
            [0., 0., -1.]
        ])
    )

    angles = func(tree, parID=0, parEND=1, ch1ID=2, ch2ID=3)
    npt.assert_allclose(angles, [0., 0., 0.5 * np.pi, -0.5 * np.pi])


def test_get_angles():

    tree = MockTree(
        np.array([
            [0., 0., 1.],
            [0., 0., 0.],
            [0., 0., -1.],
            [0., 1., 0.]
        ])
    )

    begs = [-1, 1]

    children = {1: [2, 3]}
    parents = {1: 0}

    angles = tested.PersistentAngles._get_angles(tree, begs, parents, children)
    npt.assert_allclose(angles, [[0., 0., 0., 0.], [0., 0., 0.5 * np.pi, -0.5 * np.pi]])




'''
Contains all the commonly used functions and data
useful for multiple tmd modules.
'''

term_dict = {'x': 0,
             'y': 1,
             'z': 2}

TREE_TYPE_DICT = {1: 'soma',
                  2: 'axon',
                  3: 'basal',
                  4: 'apical'}

SOMA_TYPE = 1



'''
tmd Topology algorithms implementation
'''
import numpy as np
import scipy.spatial as sp
from tmd.Topology.persistent_properties import NoProperty
from tmd.Topology.persistent_properties import PersistentAngles
from tmd.Topology.persistent_properties import PersistentMeanRadius
from tmd.Topology.analysis import sort_ph


def write_ph(ph, output_file='test.txt'):
    '''Writes a persistence diagram in
       an output file.
    '''
    wfile = open(output_file, 'w')

    for p in ph:
        wfile.write(str(p[0]) + ' ' + str(p[1]) + '\n')

    wfile.close()


def tree_to_property_barcode(tree, filtration_function, property_class=NoProperty):
    """Decompose a tree data structure into a barcode, where each bar in the barcode
    is optionally linked with a property determined by property_class.

    Args:

        filtration_function (Callable[tree] -> np.ndarray):
            The filtration function to apply on the tree

        property_class (PersistentProperty, optional): A PersistentProperty class.By
            default the NoProperty is used which does not add entries in the barcode.

    Returns:
        barcode (list): A list of bars [bar1, bar2, ..., barN], where each bar is a
            list of:
                - filtration value start
                - filtration value end
                - property_value1
                - property_value2
                - ...
                - property_valueN
    """
    point_values = filtration_function(tree)

    beg, _ = tree.sections
    parents, children = tree.parents_children

    prop = property_class(tree)

    active = tree.get_bif_term() == 0
    alives = np.where(active)[0]

    ph = []
    while len(alives) > 1:
        for alive in alives:

            p = parents[alive]
            c = children[p]

            if np.alltrue(active[c]):

                active[p] = True
                active[c] = False

                mx = np.argmax(abs(point_values[c]))
                mx_id = c[mx]

                c = np.delete(c, mx)

                for ci in c:
                    component_id = np.where(beg == p)[0][0]
                    ph.append(
                        [point_values[ci], point_values[p]] + prop.get(component_id)
                    )

                point_values[p] = point_values[mx_id]
        alives = np.where(active)[0]

    ph.append(
        [point_values[alives[0]], 0] + prop.infinite_component(beg[0])
    )  # Add the last alive component

    return ph


def _filtration_function(feature, **kwargs):
    """Returns filtration function lambda that will be applied point-wise
    on the tree"""
    return lambda tree: getattr(tree, 'get_point_' + feature)(**kwargs)


def get_persistence_diagram(tree, feature='radial_distances', **kwargs):
    '''Method to extract ph from tree that contains mutlifurcations'''
    return tree_to_property_barcode(
        tree,
        filtration_function=_filtration_function(feature, **kwargs),
        property_class=NoProperty
    )


def get_ph_angles(tree, feature='radial_distances', **kwargs):
    '''Method to extract ph from tree that contains mutlifurcations'''
    return tree_to_property_barcode(
        tree,
        filtration_function=_filtration_function(feature, **kwargs),
        property_class=PersistentAngles
    )


def get_ph_radii(tree, feature='radial_distances', **kwargs):
    """Returns the ph diagram enhanced with the corresponding encoded radii"""
    return tree_to_property_barcode(
        tree,
        filtration_function=_filtration_function(feature, **kwargs),
        property_class=PersistentMeanRadius
    )


def get_ph_neuron(neuron, feature='radial_distances', neurite_type='all', **kwargs):
    '''Method to extract ph from a neuron that contains mutlifurcations'''

    ph_all = []

    if neurite_type == 'all':
        neurite_list = ['neurites']
    else:
        neurite_list = [neurite_type]

    for t in neurite_list:
        for tr in getattr(neuron, t):
            ph_all = ph_all + get_persistence_diagram(tr, feature=feature, **kwargs)

    return ph_all


def extract_ph(tree, feature='radial_distances', output_file='test.txt',
               sort=False, **kwargs):
    '''Extracts persistent homology from tree'''
    ph = get_persistence_diagram(tree, feature=feature, **kwargs)

    if sort:
        p = sort_ph(ph)
    else:
        p = ph

    write_ph(p, output_file)


def extract_ph_neuron(neuron, feature='radial_distances', output_file=None,
                      neurite_type='all', sort=False, **kwargs):
    '''Extracts persistent homology from tree'''
    ph = get_ph_neuron(neuron, feature=feature, neurite_type='all', **kwargs)

    if sort:
        sort_ph(ph)
    else:
        p = ph

    if output_file is None:
        output_file = 'PH_' + neuron.name + '_' + neurite_type + '.txt'

    write_ph(p, output_file)


def get_lifetime(tree, feature='point_radial_distances'):
    '''Returns the sequence of birth - death times for each section.
    This can be used as the first step for the approximation of P.H.
    of the radial distances of the neuronal branches.
    '''
    begs, ends = tree.get_sections_2()
    rd = getattr(tree, 'get_' + feature)()
    lifetime = np.array(len(begs) * [np.zeros(2)])

    for i, (beg, end) in enumerate(zip(begs, ends)):
        lifetime[i] = np.array([rd[beg], rd[end]])

    return lifetime


def extract_connectivity_from_points(tree, threshold=1.0):
    '''Extract connectivity from list of points'''
    coords = np.transpose([tree.x, tree.y, tree.z])
    distances_matrix = sp.distance.cdist(coords, coords)
    mat = distances_matrix < threshold
    return mat



"""Functions for converting morphio to tmd Neuron
"""
from collections import namedtuple
import numpy as np
from tmd.Soma.Soma import Soma
from tmd.Tree import Tree


SectionData = namedtuple('SectionData', ['points', 'diameters', 'section_type', 'parents'])


def convert_morphio_soma(morphio_soma):
    """Converts a morphio's morphology
    Args:
        morphio_soma (Union[morphio.Soma])
    Returns:
        tmd_soma (Soma)
    """
    points = morphio_soma.points
    return Soma(
        x=points[:, 0],
        y=points[:, 1],
        z=points[:, 2],
        d=morphio_soma.diameters
    )


def _section_to_data(section, tree_length, start, parent):
    """Extract data from morphio section

    Args:
        section (morphio.Section): A morphio section object
        tree_length (int): Number of nodes in the tree so far

    Returns:
        tuple:
            - n (int): Number of nodes extracted
            - section_data (SectionData): namedtuple with section data
    """
    points = section.points

    # start is either 0 or 1 if the section is root or not respectively
    # dropping the first node in the latter case
    n = len(points) - start

    # each node has the previous node as a parent, except for the first
    # one, which is given from the parent
    parents = np.arange(tree_length - 1, tree_length + n - 1, dtype=np.int)
    parents[0] = parent

    return n, SectionData(
        points[start:],
        section.diameters[start:],
        int(section.type),
        parents)


def convert_morphio_trees(morphio_neuron):
    """Convert morphio morphology's trees to tmd ones

    Args:
        morphio_neuron (Union[morphio.Morphology, morphio.mut.Morphology]):
            morphio neuron object

    Yields:
        tuple:
            - tree (Tree): The constructed tmd Tree
            - tree_types (dict): The neuron tree types
    """
    total_points = len(morphio_neuron.diameters)

    x = np.empty(total_points, dtype=np.float32)
    y = np.empty(total_points, dtype=np.float32)
    z = np.empty(total_points, dtype=np.float32)
    d = np.empty(total_points, dtype=np.float32)
    t = np.empty(total_points, dtype=np.int32)
    p = np.empty(total_points, dtype=np.int64)

    section_final_nodes = np.empty(total_points, dtype=np.int)

    tree_end = 0
    for root in morphio_neuron.root_sections:

        tree_length = 0
        tree_beg = tree_end

        for section in root.iter():

            # root sections have parent -1
            if section.is_root:
                start = 0
                parent = -1
            else:
                # tmd does not use a duplicate point representation
                # thus the first point of the section is dropped
                start = 1
                parent = section_final_nodes[section.parent.id]

            n, data = _section_to_data(section, tree_length, start, parent)

            x[tree_end: n + tree_end] = data.points[:, 0]
            y[tree_end: n + tree_end] = data.points[:, 1]
            z[tree_end: n + tree_end] = data.points[:, 2]
            d[tree_end: n + tree_end] = data.diameters
            t[tree_end: n + tree_end] = data.section_type
            p[tree_end: n + tree_end] = data.parents

            tree_end += n
            tree_length += n

            # keep track of the last node in the section because we need
            # to establish the correct connectivity when we ommit the first
            # point from the children sections
            section_final_nodes[section.id] = tree_length - 1

        yield Tree.Tree(
            x=x[tree_beg: tree_end],
            y=y[tree_beg: tree_end],
            z=z[tree_beg: tree_end],
            d=d[tree_beg: tree_end],
            t=t[tree_beg: tree_end],
            p=p[tree_beg: tree_end]
        )



"""Persistent properties classes
"""
from abc import ABC, abstractmethod
import numpy as np


class PersistentProperty(ABC):
    """Abstract class for persistent properties that are defined
    on persistent components.
    """

    @abstractmethod
    def get(self, component_start):
        """Get component property"""

    @abstractmethod
    def infinite_component(self, component_start):
        """Get property for infinite component"""


class NoProperty(PersistentProperty):
    """Function class for extracting a barcode without properties
    """
    def __init__(self, _):
        pass

    def get(self, _):
        """Returns empty list, does not contribute to component"""
        return []

    def infinite_component(self, _):
        """Returns empty list, does not contribute to component"""
        return []


class PersistentMeanRadius(PersistentProperty):
    """Component mean radii
    Args:
        tree (Tree): A tree object
    """
    def __init__(self, tree):

        section_begs, section_ends = tree.sections
        self._radii = self._section_mean_radii(0.5 * tree.d, section_begs, section_ends)

    def get(self, component_start):
        """
        Args:
            component_start (int): The component start
        Returns:
            component_angles (list): A list of 1 radius
        """
        return [self._radii[component_start]]

    def infinite_component(self, component_start):
        """Returns mean radius corresponding to inf
        component
        """
        return self.get(component_start)

    @staticmethod
    def _section_mean_radii(tree_radii, section_begs, section_ends):
        """Returns the mean radius per section"""
        return np.fromiter(
            (np.mean(tree_radii[b: e]) for b, e in zip(section_begs, section_ends)),
            dtype=np.float
        )


class PersistentAngles(PersistentProperty):
    """Bifurcation angles per component
    Args:
        tree (Tree): A tree object
    """
    def __init__(self, tree):

        section_begs, _ = tree.sections
        section_parents, section_children = tree.parents_children

        self._angles = self._get_angles(
            tree, section_begs, section_parents, section_children)

    def get(self, component_start):
        """
        Args:
            component_start (int): The component start
        Returns:
            component_angles (list): A list of 4 angles
        """
        return self._angles[component_start]

    def infinite_component(self, _):
        """Given that there are not angles for the inf
        component, nans are returned.
        """
        return [np.nan, np.nan, np.nan, np.nan]

    @staticmethod
    def _phi_theta(u, v):
        """Computes the angles between vectors u, v
        in the plane x-y (phi angle) and the plane x-z (theta angle).
        Returns phi, theta

        Args:
            u (np.ndarray): (3,) First vector
            v (np.ndarray): (3,) Second vector

        Returns:
            delta_phi (float): Difference of phi_angles phi_v - phi_u
                on the x-y plane
            delta_theta (float): Difference of theta_angles th_v - th_u
                on the x-z plane
        """
        phi1 = np.arctan2(u[1], u[0])
        # pylint: disable=assignment-from-no-return
        theta1 = np.arccos(u[2] / np.linalg.norm(u))

        # pylint: disable=assignment-from-no-return
        phi2 = np.arctan2(v[1], v[0])
        theta2 = np.arccos(v[2] / np.linalg.norm(v))

        delta_phi = phi2 - phi1  # np.abs(phi1 - phi2)
        delta_theta = theta2 - theta1  # np.abs(theta1 - theta2)

        return delta_phi, delta_theta  # dphi, dtheta

    @staticmethod
    def _angles_tree(tree, parID, parEND, ch1ID, ch2ID):
        '''Computes the x-y and x-z angles between parent
           and children within the given tree.

        Args:
            tree (Tree): Morphology tree
            parID (int): Id of parent section
            parEnd (int): Id of parent section end
            ch1ID (int): ID of first child
            ch2ID (int): ID of section child

        Returns:
            list:
                dphi (float):
                    Absolute difference of phi_angles between parent and first child
                dtheta (float):
                    Absolute difference of theta_angles between parent and first child
                delta_phi (float): Difference of phi_angles phi_v - phi_u
                    on the x-y plane
                delta_theta (float): Difference of theta_angles th_v - th_u
                    on the x-z plane
        '''

        parent_direction = tree.get_direction_between(start_id=parID, end_id=parEND)
        child1_direction = tree.get_direction_between(start_id=parEND, end_id=ch1ID)
        child2_direction = tree.get_direction_between(start_id=parEND, end_id=ch2ID)

        phi1, theta1 = PersistentAngles._phi_theta(parent_direction, child1_direction)
        phi2, theta2 = PersistentAngles._phi_theta(parent_direction, child2_direction)

        if np.abs(phi1) < np.abs(phi2):
            dphi = phi1
            dtheta = theta1
            delta_phi, delta_theta = PersistentAngles._phi_theta(child1_direction, child2_direction)
        else:
            dphi = phi2
            dtheta = theta2
            delta_phi, delta_theta = PersistentAngles._phi_theta(child2_direction, child1_direction)

        return [dphi, dtheta, delta_phi, delta_theta]

    @staticmethod
    def _get_angles(tree, beg, parents, children):
        """Returns the angles between all the triplets (parent, child1, child2)
        of the tree"""
        angles = [[0, 0, 0, 0], ]  # Null angle for non bif point

        for b in beg[1:]:

            angleBetween = PersistentAngles._angles_tree(
                tree, parID=parents[b], parEND=b, ch1ID=children[b][0], ch2ID=children[b][1])

            angles.append(angleBetween)

        return angles



import os
from collections import namedtuple
import mock
import numpy as np
from numpy import testing as npt
import morphio
from tmd.Neuron.Neuron import Neuron
from tmd.io import conversion as tested
from tmd.io.io import load_neuron
from tmd.io.io import load_neuron_from_morphio

_path = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(_path, 'data')


MockSection = namedtuple('MockSection', ['id', 'points', 'diameters', 'type', 'parent'])


class MockSection:

    def __init__(self, id, points, diameters, type, parent=None):

        self.id = id
        self.points = points
        self.diameters = diameters
        self.type = type
        self.parent = parent
        self.traversal = []

    def iter(self):
        return self.traversal

    @property
    def is_root(self):
        return self.parent is None


class MockNeuron:

    def __init__(self):

        root = MockSection(
                id=0,
                points = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]),
                diameters = np.array([0.1, 0.2, 0.3]),
                type=2,
                parent=None
        )

        child1 = MockSection(
                id=1,
                points = np.array([[0.7, 0.8, 0.9], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]),
                diameters = np.array([0.3, 0.4, 0.5]),
                type=2,
                parent=root
        )

        child2 = MockSection(
                id=2,
                points = np.array([[0.7, 0.8, 0.9], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]]),
                diameters = np.array([0.3, 0.6, 0.7]),
                type=2,
                parent=root
        )

        root.traversal = [root, child1, child2]

        self.root_sections = [root]

    @property
    def diameters(self):
        return np.hstack([s.diameters for root in self.root_sections for s in root.iter()])


def test_convert_morphio_soma():

    morphio_soma = mock.Mock(
        points = np.array([
            [0., 1., 2.],
            [2., 3., 4.]
        ]),
        diameters=np.array([2.1, 3.4])
    )

    soma = tested.convert_morphio_soma(morphio_soma)

    npt.assert_allclose(soma.x, [0., 2.])
    npt.assert_allclose(soma.y, [1., 3.])
    npt.assert_allclose(soma.z, [2., 4.])
    npt.assert_allclose(soma.d, [2.1, 3.4])


def test_section_to_data():


    section = MockSection(
        id=0,
        points=np.array([
            [0.1, 0.2, 0.3],
            [0.4, 0.5, 0.6],
            [0.7, 0.8, 0.9]
        ]),
        diameters=np.array([1.2, 1.3, 1.4]),
        type=3,
        parent=None
    )

    n, data = tested._section_to_data(section, tree_length=11, start=0, parent=-1)

    npt.assert_equal(n, 3)
    npt.assert_allclose(data.points, section.points)
    npt.assert_allclose(section.diameters, data.diameters)
    npt.assert_equal(data.section_type, 3)
    npt.assert_array_equal(data.parents, [-1, 11 + 0, 11 + 1])

    n, data = tested._section_to_data(section, tree_length=2, start=1, parent=5)

    npt.assert_equal(n, 2)
    npt.assert_allclose(data.points, section.points[1:])
    npt.assert_allclose(data.diameters, section.diameters[1:])
    npt.assert_equal(data.section_type, 3)
    npt.assert_array_equal(data.parents, [5, 2 + 0])


def test_convert_morphio_trees():

    morphio_neuron = MockNeuron()

    trees = list(tested.convert_morphio_trees(morphio_neuron))

    assert len(trees) == 1

    tree = trees[0]

    npt.assert_allclose(tree.x, [0.1, 0.4, 0.7, 0.4, 0.7, 0.4, 0.7])
    npt.assert_allclose(tree.y, [0.2, 0.5, 0.8, 0.5, 0.8, 0.5, 0.8])
    npt.assert_allclose(tree.z, [0.3, 0.6, 0.9, 0.6, 0.9, 0.6, 0.9])
    npt.assert_allclose(tree.d, [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
    npt.assert_array_equal(tree.t, [2, 2, 2, 2, 2, 2, 2])
    npt.assert_array_equal(tree.p, [-1, 0, 1, 2, 3, 2, 5])


def _assert_neurons_equal(neuron1, neuron2):

    npt.assert_allclose(neuron1.soma.x, neuron2.soma.x)
    npt.assert_allclose(neuron1.soma.y, neuron2.soma.y)
    npt.assert_allclose(neuron1.soma.z, neuron2.soma.z)
    npt.assert_allclose(neuron1.soma.d, neuron2.soma.d)

    for neurite1, neurite2 in zip(neuron1.neurites, neuron2.neurites):

        npt.assert_allclose(neurite1.x, neurite2.x)
        npt.assert_allclose(neurite1.y, neurite2.y)
        npt.assert_allclose(neurite1.z, neurite2.z)
        npt.assert_allclose(neurite1.d, neurite2.d)
        npt.assert_array_equal(neurite1.t, neurite2.t)
        npt.assert_array_equal(neurite1.p, neurite2.p)


def test_neuron_building_consistency__h5():

    path = f'{DATA_PATH}/valid/C010398B-P2.h5'

    neuron1 = load_neuron(path)
    neuron2 = load_neuron_from_morphio(path)

    _assert_neurons_equal(neuron1, neuron2)

    neuron2 = load_neuron_from_morphio(morphio.Morphology(path))

    _assert_neurons_equal(neuron1, neuron2)


def test_neuron_building_consistency__swc():

    path = f'{DATA_PATH}/valid/C010398B-P2.CNG.swc'

    neuron1 = load_neuron(path)
    neuron2 = load_neuron_from_morphio(path)

    _assert_neurons_equal(neuron1, neuron2)

    neuron2 = load_neuron_from_morphio(morphio.Morphology(path))

    _assert_neurons_equal(neuron1, neuron2)


